/*
 * Copyright (C) 2008-2009 the original author or authors
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *         http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.google.code.liquidform;

import java.util.Collection;
import java.util.Date;

import com.google.code.liquidform.internal.AliasFactory;
import com.google.code.liquidform.internal.FrameworkObjectFactory;
import com.google.code.liquidform.internal.managedobjects.FrameworkManagedObject;

/**
 * Provides several constructs for recording the use of standard JPA functions.
 * The function names here should be self explanatory and are borrowed from the
 * JPA query syntax.
 * 
 * <p>
 * The return types of those functions is specified by section 4.6.16.1 thru
 * 4.6.16.3 of JSR220
 * </p>
 */
public class Functions {

	private Functions() {
		// utility class
	}

	// Functions returning Strings (4.6.16.1)
	public static String concat(final String lhs, final String rhs) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "concat(" + AliasFactory.safeStringForm(lhs) + ", "
						+ AliasFactory.safeStringForm(rhs) + ")";
			}
		});
		return s;
	}

	public static String substring(final String string, final Integer start,
			final Integer length) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "substring(" + AliasFactory.safeStringForm(string)
						+ ", " + AliasFactory.safeStringForm(start) + ", "
						+ AliasFactory.safeStringForm(length) + ")";
			}
		});
		return s;
	}

	public static String trim(final String string) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "trim(" + AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	/**
	 * Whether to trim at left, right or both ends.
	 */
	public static enum TrimSpecification {
		LEADING("leading "), TRAILING("trailing "), BOTH("both ");

		private final String stringRepresentation;

		private TrimSpecification(String stringRepresentation) {
			this.stringRepresentation = stringRepresentation;
		}
	}

	public static final TrimSpecification LEADING = TrimSpecification.LEADING;
	public static final TrimSpecification TRAILING = TrimSpecification.TRAILING;
	public static final TrimSpecification BOTH = TrimSpecification.BOTH;

	public static String trim(final String string, final Character toTrim) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "trim(" + AliasFactory.safeStringForm(toTrim) + " from "
						+ AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	public static String trim(final String string,
			final TrimSpecification trimSpecification) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "trim(" + trimSpecification.stringRepresentation
						+ " from " + AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	public static String trim(final String string, final Character toTrim,
			final TrimSpecification trimSpecification) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "trim(" + trimSpecification.stringRepresentation
						+ AliasFactory.safeStringForm(toTrim) + " from "
						+ AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	public static String lower(final String string) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "lower(" + AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	public static String upper(final String string) {
		String s = FrameworkObjectFactory.INSTANCE.makeNew(String.class);
		AliasFactory.remember(s, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "upper(" + AliasFactory.safeStringForm(string) + ")";
			}
		});
		return s;
	}

	// Functions returning numerics
	public static Integer length(final String string) {
		Integer i = FrameworkObjectFactory.INSTANCE.makeNew(Integer.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "length(" + AliasFactory.safeStringForm(string) + ")";
			}
		});
		return i;
	}

	public static Integer locate(final String string, final String in) {
		Integer i = FrameworkObjectFactory.INSTANCE.makeNew(Integer.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "locate(" + AliasFactory.safeStringForm(string) + ", "
						+ AliasFactory.safeStringForm(in) + ")";
			}
		});
		return i;
	}

	public static Integer locate(final String string, final String in,
			final Integer start) {
		Integer i = FrameworkObjectFactory.INSTANCE.makeNew(Integer.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "locate(" + AliasFactory.safeStringForm(string) + ", "
						+ AliasFactory.safeStringForm(in) + ", "
						+ AliasFactory.safeStringForm(start) + ")";
			}
		});
		return i;
	}

	public static <N extends Number> N abs(final N expr) {
		N i = FrameworkObjectFactory.INSTANCE.makeNew(expr);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "abs(" + AliasFactory.safeStringForm(expr) + ")";
			}
		});
		return i;
	}

	public static Double sqrt(final Number expr) {
		Double i = FrameworkObjectFactory.INSTANCE.makeNew(Double.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "sqrt(" + AliasFactory.safeStringForm(expr) + ")";
			}
		});
		return i;
	}

	public static Integer mod(final Integer expr, final Integer divider) {
		Integer i = FrameworkObjectFactory.INSTANCE.makeNew(Integer.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "mod(" + AliasFactory.safeStringForm(expr) + ", "
						+ AliasFactory.safeStringForm(divider) + ")";
			}
		});
		return i;
	}

	public static Integer size(final Collection<?> collection) {
		Integer i = FrameworkObjectFactory.INSTANCE.makeNew(Integer.class);
		AliasFactory.remember(i, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "size(" + AliasFactory.safeStringForm(collection) + ")";
			}
		});
		return i;
	}

	// functions returning dates
	public static Date currentDate() {
		Date d = FrameworkObjectFactory.INSTANCE.makeNew(Date.class);
		AliasFactory.remember(d, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "current_date";
			}
		});
		return d;
	}

	public static Date currentTime() {
		Date d = FrameworkObjectFactory.INSTANCE.makeNew(Date.class);
		AliasFactory.remember(d, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "current_time";
			}
		});
		return d;
	}

	public static Date currentTimestamp() {
		Date d = FrameworkObjectFactory.INSTANCE.makeNew(Date.class);
		AliasFactory.remember(d, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "current_timestamp";
			}
		});
		return d;
	}

}
