/*
 * Copyright (C) 2008-2009 the original author or authors
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *         http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.code.liquidform;

import java.util.Collection;
import java.util.Date;

import javax.persistence.Entity;

import net.sf.cglib.proxy.MethodInterceptor;

import com.google.code.liquidform.conditions.AllOrAny;
import com.google.code.liquidform.conditions.AllOrAnyOperator;
import com.google.code.liquidform.conditions.BetweenOperator;
import com.google.code.liquidform.conditions.BinaryExpression;
import com.google.code.liquidform.conditions.ComparisonOperator;
import com.google.code.liquidform.conditions.ConditionalExpression;
import com.google.code.liquidform.conditions.EqualityOperator;
import com.google.code.liquidform.conditions.LikeOperator;
import com.google.code.liquidform.conditions.TernaryExpression;
import com.google.code.liquidform.internal.AliasFactory;
import com.google.code.liquidform.internal.CglibProxyFactory;
import com.google.code.liquidform.internal.FrameworkObjectFactory;
import com.google.code.liquidform.internal.ObjectUtils;
import com.google.code.liquidform.internal.PropertyAccessRecordingInvocationHandler;
import com.google.code.liquidform.internal.managedobjects.CollectionAssociationPath;
import com.google.code.liquidform.internal.managedobjects.FrameworkManagedObject;

/**
 * This is the main entry point to the framework, providing the ability to alias
 * a type to a local variable and bootstrapping a query through the
 * <tt>SELECT x, y, z</tt> construct.
 */
public final class LiquidForm {

	private static MethodInterceptor interceptor = new PropertyAccessRecordingInvocationHandler();

	private LiquidForm() {
		// utility class
	}

	/**
	 * Declare a new <i>identification variable</i> typed as the <i>abstract
	 * schema type</i> of the given persistent entity type <tt>clazz</tt>. Used
	 * to assign an alias to a strongly typed local variable that will later be
	 * used in the query. Typical use looks like this
	 * 
	 * <pre>
	 * final Person p = LiquidForm.use(Person.class, &quot;p&quot;);
	 * String q = select(p.getName()).from(Person.class).as(p); //etc..
	 * </pre>
	 * 
	 * @param <T>
	 *            the type of persistent entity that will be aliased
	 * @param clazz
	 *            the type of persistent entity that will be aliased
	 * @param alias
	 *            is necessary to inform the framework of the alias name. This
	 *            should typically be the name chosen for the local variable.
	 * @return
	 */
	public static <T> T alias(Class<T> clazz, String alias) {

		T asn = CglibProxyFactory.INSTANCE.createProxy(clazz.getClassLoader(),
				interceptor, clazz);

		String entityType = inferEntityName(clazz);
		AliasFactory.rememberAsAbstractSchemaName(asn, entityType, alias);

		return asn;

	}

	/**
	 * Declare a new <i>identification variable</i> as an alias to a property of
	 * a persistent object that was created by a call to
	 * {@link #alias(Class, String)}. This is useful for example when you
	 * perform joins in your query, or if you need an alias to select some
	 * property.
	 * 
	 * <pre>
	 * final Person p = LiquidForm.use(Person.class, &quot;p&quot;);
	 * final Address a = LiquidForm.alias(p.getAddress(), &quot;a&quot;);
	 * select(p, a).from(Person.class).as(p).innerJoin(p.getAddress()).as(a); // etc
	 * </pre>
	 * 
	 * @param <T>
	 *            the type that is being aliased
	 * @param property
	 *            should be a property retrieval on an instance returned by
	 *            {@link #alias(Class, String)}
	 * @param alias
	 *            is necessary to inform the framework of the alias name. This
	 *            should typically be the name chosen for the local variable.
	 * @return
	 */
	public static <T> T alias(T property, String alias) {
		// PropertyAccessRecordingInvocationHandler has already recorded some
		// info about this property access
		// Add the alias to it
		String path = AliasFactory.about(property).getTargetRepresentation();
		AliasFactory.rememberAsAliasedAssociationPath(property, path, alias);
		return property;
	}

	/**
	 * Declare a new <i>identification variable</i> as an alias to a member of
	 * the persistent collection wrapped in a call to {@link #in(Collection)}.
	 * 
	 * <pre>
	 * Order o = alias(Order.class, &quot;o&quot;);
	 * LineItem l = alias(in(o.getLineItems()), &quot;l&quot;);
	 * select(distinct(l.getProduct())).from(Order.class).as(o)
	 *     .andFrom(in(o.getLineItems())).as(l)
	 * </pre>
	 */
	@SuppressWarnings("unchecked")
	public static <T, U extends T> U alias(InCollection<T> inCollection,
			String alias) {
		// PropertyAccessRecordingInvocationHandler has already recorded some
		// info about this property access
		// Add the alias to it
		CollectionAssociationPath<U> collectionAssociationPath = (CollectionAssociationPath<U>) AliasFactory
				.about(inCollection.getCollection());
		String path = AliasFactory.about(inCollection.getCollection())
				.getTargetRepresentation();

		U elementProxy = collectionAssociationPath.getElementProxy();
		AliasFactory.rememberAsAliasedInCollection(elementProxy, path, alias);
		return elementProxy;
	}

	/**
	 * Declare a new <i>identification variable</i> as an alias to a member of
	 * the persistent collection.
	 * 
	 * <pre>
	 * Order o = alias(Order.class, &quot;o&quot;);
	 * LineItem l = alias(o.getLineItems(), &quot;l&quot;);
	 * select(distinct(o)).from(Order.class).as(o)
	 *     .innerJoin(o.getLineItems()).as(l)
	 * </pre>
	 */
	@SuppressWarnings("unchecked")
	public static <T, U extends T> U alias(Collection<T> collection,
			String alias) {
		// PropertyAccessRecordingInvocationHandler has already recorded some
		// info about this property access
		// Add the alias to it
		CollectionAssociationPath<U> collectionAssociationPath = (CollectionAssociationPath<U>) AliasFactory
				.about(collection);
		String path = AliasFactory.about(collection).getTargetRepresentation();

		U elementProxy = collectionAssociationPath.getElementProxy();
		AliasFactory
				.rememberAsAliasedAssociationPath(elementProxy, path, alias);
		return elementProxy;
	}

	/**
	 * Used in conjunction with {@link #alias(InCollection, String)}, prepare a
	 * <i>collection member declaration</i>. This method invocation is similar
	 * in usage to the <tt>IN</tt> keyword in JPQL.
	 */
	public static <T> InCollection<T> in(Collection<T> collection) {
		if (AliasFactory.about(collection) == null) {
			throw new IllegalStateException();
		}
		return new InCollection<T>(collection);
	}

	/**
	 * If the persistent class bears the Entity(name() = "name") annotation
	 * parameter, use this as the entity name. Else, simply use the class'
	 * simple name.
	 */
	private static String inferEntityName(Class<?> clazz) {
		Entity entityAnn = clazz.getAnnotation(Entity.class);
		if (entityAnn == null) {
			throw new IllegalArgumentException(clazz
					+ " is not a JPA persistent class");
		}
		return "".equals(entityAnn.name()) ? clazz.getSimpleName() : entityAnn
				.name();
	}

	/**
	 * Used to bootstrap a query creation by specifying what is to be selected.
	 * This is the direct equivalent of JPA's <tt>SELECT x, y, z</tt> construct.
	 * <p>
	 * <b>Note : </b> The funny signature is here to force at least one object
	 * to select
	 * </p>
	 * 
	 * @param prop
	 *            the objects that will be selected
	 * @param more
	 *            the objects that will be selected
	 * @return an object that must be used to construct the query
	 * @see SelectClause#from(Class)
	 */
	public static SelectClause<Object[]> select(Object prop, Object... more) {
		return new SelectClause<Object[]>(ObjectUtils.merge(Object.class, prop,
				more));
	}

	/**
	 * Used to bootstrap a query creation by specifying what is to be selected.
	 * This is the direct equivalent of JPA's <tt>SELECT x</tt> construct.
	 * <p>
	 * 
	 * @param prop
	 *            the objects that will be selected
	 * @return an object that must be used to construct the query
	 * @see SelectClause#from(Class)
	 */
	public static <S> SelectClause<S> select(S prop) {
		return new SelectClause<S>(new Object[] { prop });
	}

	/**
	 * Optionally wrap a selected identification variable with the
	 * <tt>object()</tt> operator.
	 * <p>
	 * Note : <tt>SELECT x</tt> and <tt>SELECT OBJECT(x)</tt> are equivalent.
	 * </p>
	 */
	public static <V> V object(final V identificationVariable) {
		// TODO : check that argument is not a path expression
		V v = FrameworkObjectFactory.INSTANCE.makeNew(identificationVariable);
		AliasFactory.remember(v, new FrameworkManagedObject() {
			@Override
			public String getTargetRepresentation() {
				return "object("
						+ AliasFactory.safeStringForm(identificationVariable)
						+ ")";
			}
		});
		return v;
	}

	// TODO : should we try to avoid side effects when calling the constructor ?
	// if not, how can we leverage the compiler type checking (ie allow users
	// to write "new MyPojo(p.getName(), p.getAge())"
	/*
	 * public static <V> V constr(final Class<V> clazz, final Object... args) {
	 * V v = FrameworkObjectFactory.INSTANCE.makeNew(clazz);
	 * AliasFactory.remember(v, new FrameworkManagedObject() {
	 * 
	 * @Override public String getTargetRepresentation() { StringBuilder sb =
	 * new StringBuilder("new ").append( clazz.getCanonicalName()).append('(');
	 * for (int i = 0; i < args.length; i++) {
	 * sb.append(AliasFactory.safeStringForm(args[i])); if (i < args.length + 1)
	 * { sb.append(", "); } } return sb.append(')').toString(); } }); return v;
	 * }
	 */

	// ====================== S O R T D I R E C T I O N S ======================
	public static final SortDirection ASC = SortDirection.ASC;

	public static final SortDirection DESC = SortDirection.DESC;

	// ==== C O N D I T I O N A L _ E X P R E S S I O N _ F A C T O R I E S ====

	// For entities, this makes sure that both operands have a inheritance
	// relationship ...
	// TODO : this is wrong, see 4.12 V must be the same as W

	// entity {=|<>} entity
	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression eq(V lhs, W rhs) {
		return new BinaryExpression<V, W>(lhs, EqualityOperator.EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression neq(V lhs, W rhs) {
		return new BinaryExpression<V, W>(lhs, EqualityOperator.NOT_EQUALS, rhs);
	}

	// entity {=|<>} all_or_any_expression
	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression eq(V lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<V, AllOrAny<W>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression neq(V lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<V, AllOrAny<W>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	// ... hence special cases need explicit disambiguation
	// arithmetic {=|<>} arithmetic
	public static ConditionalExpression eq(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression eq(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression eq(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression neq(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	public static ConditionalExpression neq(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	public static ConditionalExpression neq(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	// date {=|<>} date
	public static ConditionalExpression eq(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs, EqualityOperator.EQUALS,
				rhs);
	}

	public static ConditionalExpression eq(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression eq(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression neq(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	public static ConditionalExpression neq(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	public static ConditionalExpression neq(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	// arithmetic {=|<>} all_or_any_expression
	public static ConditionalExpression eq(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression neq(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	// date {=|<>} all_or_any_expression
	public static ConditionalExpression eq(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	public static ConditionalExpression neq(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	// Comparison Date
	// date {<|<=|>|>=} date
	public static ConditionalExpression lt(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs, ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs, ComparisonOperator.LTE,
				rhs);
	}

	public static ConditionalExpression gt(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs, ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Date lhs, Date rhs) {
		return new BinaryExpression<Date, Date>(lhs, ComparisonOperator.GTE,
				rhs);
	}

	public static ConditionalExpression lt(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Date lhs,
			SubQuery<? extends Date> rhs) {
		return new BinaryExpression<Date, SubQuery<? extends Date>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(SubQuery<? extends Date> lhs,
			Date rhs) {
		return new BinaryExpression<SubQuery<? extends Date>, Date>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// date {<|<=|>|>=} all_or_any_expression
	public static ConditionalExpression lt(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Date lhs,
			AllOrAny<? extends Date> rhs) {
		return new BinaryExpression<Date, AllOrAny<? extends Date>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// date between date and date
	public static ConditionalExpression between(Date lhs, Date rhs1, Date rhs2) {
		return new TernaryExpression<Date, Date, Date>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Date lhs, Date rhs1,
			Date rhs2) {
		return new TernaryExpression<Date, Date, Date>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	// between signatures with SubQuery<> params

	// TODO : think about sig with 3 SQ<> : what relation between U, V and W?
	public static ConditionalExpression between(Number lhs, Number rhs1,
			SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<Number, Number, SubQuery<? extends Number>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Number lhs,
			SubQuery<? extends Number> rhs1, Number rhs2) {
		return new TernaryExpression<Number, SubQuery<? extends Number>, Number>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Number lhs,
			SubQuery<? extends Number> rhs1, SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<Number, SubQuery<? extends Number>, SubQuery<? extends Number>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Number> lhs,
			Number rhs1, Number rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, Number, Number>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Number> lhs,
			Number rhs1, SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, Number, SubQuery<? extends Number>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Number> lhs,
			SubQuery<? extends Number> rhs1, Number rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, SubQuery<? extends Number>, Number>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Date lhs, Date rhs1,
			SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<Date, Date, SubQuery<? extends Date>>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Date lhs,
			SubQuery<? extends Date> rhs1, Date rhs2) {
		return new TernaryExpression<Date, SubQuery<? extends Date>, Date>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Date lhs,
			SubQuery<? extends Date> rhs1, SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<Date, SubQuery<? extends Date>, SubQuery<? extends Date>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Date> lhs,
			Date rhs1, Date rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, Date, Date>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Date> lhs,
			Date rhs1, SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, Date, SubQuery<? extends Date>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<? extends Date> lhs,
			SubQuery<? extends Date> rhs1, Date rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, SubQuery<? extends Date>, Date>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Character lhs, Character rhs1,
			SubQuery<Character> rhs2) {
		return new TernaryExpression<Character, Character, SubQuery<Character>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Character lhs,
			SubQuery<Character> rhs1, Character rhs2) {
		return new TernaryExpression<Character, SubQuery<Character>, Character>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(Character lhs,
			SubQuery<Character> rhs1, SubQuery<Character> rhs2) {
		return new TernaryExpression<Character, SubQuery<Character>, SubQuery<Character>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<Character> lhs,
			Character rhs1, Character rhs2) {
		return new TernaryExpression<SubQuery<Character>, Character, Character>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<Character> lhs,
			Character rhs1, SubQuery<Character> rhs2) {
		return new TernaryExpression<SubQuery<Character>, Character, SubQuery<Character>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<Character> lhs,
			SubQuery<Character> rhs1, Character rhs2) {
		return new TernaryExpression<SubQuery<Character>, SubQuery<Character>, Character>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(String lhs, String rhs1,
			SubQuery<String> rhs2) {
		return new TernaryExpression<String, String, SubQuery<String>>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(String lhs,
			SubQuery<String> rhs1, String rhs2) {
		return new TernaryExpression<String, SubQuery<String>, String>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(String lhs,
			SubQuery<String> rhs1, SubQuery<String> rhs2) {
		return new TernaryExpression<String, SubQuery<String>, SubQuery<String>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<String> lhs,
			String rhs1, String rhs2) {
		return new TernaryExpression<SubQuery<String>, String, String>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<String> lhs,
			String rhs1, SubQuery<String> rhs2) {
		return new TernaryExpression<SubQuery<String>, String, SubQuery<String>>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression between(SubQuery<String> lhs,
			SubQuery<String> rhs1, String rhs2) {
		return new TernaryExpression<SubQuery<String>, SubQuery<String>, String>(
				lhs, BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	// notBeetween with SubQuery<>
	// TODO : 3 SQ<> version
	public static ConditionalExpression notBetween(Number lhs, Number rhs1,
			SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<Number, Number, SubQuery<? extends Number>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Number lhs,
			SubQuery<? extends Number> rhs1, Number rhs2) {
		return new TernaryExpression<Number, SubQuery<? extends Number>, Number>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Number lhs,
			SubQuery<? extends Number> rhs1, SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<Number, SubQuery<? extends Number>, SubQuery<? extends Number>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Number> lhs, Number rhs1, Number rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, Number, Number>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Number> lhs, Number rhs1,
			SubQuery<? extends Number> rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, Number, SubQuery<? extends Number>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Number> lhs, SubQuery<? extends Number> rhs1,
			Number rhs2) {
		return new TernaryExpression<SubQuery<? extends Number>, SubQuery<? extends Number>, Number>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Date lhs, Date rhs1,
			SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<Date, Date, SubQuery<? extends Date>>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Date lhs,
			SubQuery<? extends Date> rhs1, Date rhs2) {
		return new TernaryExpression<Date, SubQuery<? extends Date>, Date>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Date lhs,
			SubQuery<? extends Date> rhs1, SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<Date, SubQuery<? extends Date>, SubQuery<? extends Date>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Date> lhs, Date rhs1, Date rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, Date, Date>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Date> lhs, Date rhs1,
			SubQuery<? extends Date> rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, Date, SubQuery<? extends Date>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(
			SubQuery<? extends Date> lhs, SubQuery<? extends Date> rhs1,
			Date rhs2) {
		return new TernaryExpression<SubQuery<? extends Date>, SubQuery<? extends Date>, Date>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Character lhs,
			Character rhs1, SubQuery<Character> rhs2) {
		return new TernaryExpression<Character, Character, SubQuery<Character>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Character lhs,
			SubQuery<Character> rhs1, Character rhs2) {
		return new TernaryExpression<Character, SubQuery<Character>, Character>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Character lhs,
			SubQuery<Character> rhs1, SubQuery<Character> rhs2) {
		return new TernaryExpression<Character, SubQuery<Character>, SubQuery<Character>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<Character> lhs,
			Character rhs1, Character rhs2) {
		return new TernaryExpression<SubQuery<Character>, Character, Character>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<Character> lhs,
			Character rhs1, SubQuery<Character> rhs2) {
		return new TernaryExpression<SubQuery<Character>, Character, SubQuery<Character>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<Character> lhs,
			SubQuery<Character> rhs1, Character rhs2) {
		return new TernaryExpression<SubQuery<Character>, SubQuery<Character>, Character>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(String lhs, String rhs1,
			SubQuery<String> rhs2) {
		return new TernaryExpression<String, String, SubQuery<String>>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(String lhs,
			SubQuery<String> rhs1, String rhs2) {
		return new TernaryExpression<String, SubQuery<String>, String>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(String lhs,
			SubQuery<String> rhs1, SubQuery<String> rhs2) {
		return new TernaryExpression<String, SubQuery<String>, SubQuery<String>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<String> lhs,
			String rhs1, String rhs2) {
		return new TernaryExpression<SubQuery<String>, String, String>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<String> lhs,
			String rhs1, SubQuery<String> rhs2) {
		return new TernaryExpression<SubQuery<String>, String, SubQuery<String>>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(SubQuery<String> lhs,
			SubQuery<String> rhs1, String rhs2) {
		return new TernaryExpression<SubQuery<String>, SubQuery<String>, String>(
				lhs, BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	// Comparison Number
	// arithmetic {<|<=|>|>=} arithmetic
	public static ConditionalExpression lt(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs, ComparisonOperator.LT,
				rhs);
	}

	public static ConditionalExpression lte(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs, ComparisonOperator.GT,
				rhs);
	}

	public static ConditionalExpression gte(Number lhs, Number rhs) {
		return new BinaryExpression<Number, Number>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Number lhs,
			SubQuery<? extends Number> rhs) {
		return new BinaryExpression<Number, SubQuery<? extends Number>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(SubQuery<? extends Number> lhs,
			Number rhs) {
		return new BinaryExpression<SubQuery<? extends Number>, Number>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// arithmetic {<|<=|>|>=} all_or_any_expression
	public static ConditionalExpression lt(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Number lhs,
			AllOrAny<? extends Number> rhs) {
		return new BinaryExpression<Number, AllOrAny<? extends Number>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// arithmetic between arithmetic
	public static ConditionalExpression between(Number lhs, Number rhs1,
			Number rhs2) {
		return new TernaryExpression<Number, Number, Number>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Number lhs, Number rhs1,
			Number rhs2) {
		return new TernaryExpression<Number, Number, Number>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	// Comparison String
	// string {<|<=|>|>=} string
	public static ConditionalExpression lt(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs, ComparisonOperator.LT,
				rhs);
	}

	public static ConditionalExpression lte(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs, ComparisonOperator.GT,
				rhs);
	}

	public static ConditionalExpression gte(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(String lhs, SubQuery<String> rhs) {
		return new BinaryExpression<String, SubQuery<String>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(String lhs, SubQuery<String> rhs) {
		return new BinaryExpression<String, SubQuery<String>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(String lhs, SubQuery<String> rhs) {
		return new BinaryExpression<String, SubQuery<String>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(String lhs, SubQuery<String> rhs) {
		return new BinaryExpression<String, SubQuery<String>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// string {<|<=|>|>=} all_or_any_expression
	public static ConditionalExpression lt(String lhs, AllOrAny<String> rhs) {
		return new BinaryExpression<String, AllOrAny<String>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(String lhs, AllOrAny<String> rhs) {
		return new BinaryExpression<String, AllOrAny<String>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(String lhs, AllOrAny<String> rhs) {
		return new BinaryExpression<String, AllOrAny<String>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(String lhs, AllOrAny<String> rhs) {
		return new BinaryExpression<String, AllOrAny<String>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// string between string and string
	public static ConditionalExpression between(String lhs, String rhs1,
			String rhs2) {
		return new TernaryExpression<String, String, String>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(String lhs, String rhs1,
			String rhs2) {
		return new TernaryExpression<String, String, String>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	// Comparison Character
	// char {<|<=|>|>=} char
	public static ConditionalExpression lt(Character lhs, Character rhs) {
		return new BinaryExpression<Character, Character>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Character lhs, Character rhs) {
		return new BinaryExpression<Character, Character>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Character lhs, Character rhs) {
		return new BinaryExpression<Character, Character>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Character lhs, Character rhs) {
		return new BinaryExpression<Character, Character>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(Character lhs,
			SubQuery<Character> rhs) {
		return new BinaryExpression<Character, SubQuery<Character>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Character lhs,
			SubQuery<Character> rhs) {
		return new BinaryExpression<Character, SubQuery<Character>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Character lhs,
			SubQuery<Character> rhs) {
		return new BinaryExpression<Character, SubQuery<Character>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Character lhs,
			SubQuery<Character> rhs) {
		return new BinaryExpression<Character, SubQuery<Character>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	public static ConditionalExpression lt(SubQuery<Character> lhs,
			Character rhs) {
		return new BinaryExpression<SubQuery<Character>, Character>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(SubQuery<Character> lhs,
			Character rhs) {
		return new BinaryExpression<SubQuery<Character>, Character>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(SubQuery<Character> lhs,
			Character rhs) {
		return new BinaryExpression<SubQuery<Character>, Character>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(SubQuery<Character> lhs,
			Character rhs) {
		return new BinaryExpression<SubQuery<Character>, Character>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// char {<|<=|>|>=} all_or_any_expression
	public static ConditionalExpression lt(Character lhs,
			AllOrAny<Character> rhs) {
		return new BinaryExpression<Character, AllOrAny<Character>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	public static ConditionalExpression lte(Character lhs,
			AllOrAny<Character> rhs) {
		return new BinaryExpression<Character, AllOrAny<Character>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	public static ConditionalExpression gt(Character lhs,
			AllOrAny<Character> rhs) {
		return new BinaryExpression<Character, AllOrAny<Character>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	public static ConditionalExpression gte(Character lhs,
			AllOrAny<Character> rhs) {
		return new BinaryExpression<Character, AllOrAny<Character>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// char between char and char
	public static ConditionalExpression between(Character lhs, Character rhs1,
			Character rhs2) {
		return new TernaryExpression<Character, Character, Character>(lhs,
				BetweenOperator.BETWEEN, rhs1, rhs2);
	}

	public static ConditionalExpression notBetween(Character lhs,
			Character rhs1, Character rhs2) {
		return new TernaryExpression<Character, Character, Character>(lhs,
				BetweenOperator.NOT_BETWEEN, rhs1, rhs2);
	}

	// like String
	// TODO : 4.6.9 : rhs(=pattern) should be literal or input param
	public static ConditionalExpression like(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs, LikeOperator.LIKE, rhs);
	}

	public static ConditionalExpression like(String lhs, String rhs,
			Character escape) {
		return new TernaryExpression<String, String, Character>(lhs,
				LikeOperator.LIKE, rhs, escape);
	}

	public static ConditionalExpression notLike(String lhs, String rhs) {
		return new BinaryExpression<String, String>(lhs, LikeOperator.NOT_LIKE,
				rhs);
	}

	public static ConditionalExpression notLike(String lhs, String rhs,
			Character escape) {
		return new TernaryExpression<String, String, Character>(lhs,
				LikeOperator.NOT_LIKE, rhs, escape);
	}

	public static ConditionalExpression like(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				LikeOperator.LIKE, rhs);
	}

	public static ConditionalExpression like(SubQuery<String> lhs, String rhs,
			Character escape) {
		return new TernaryExpression<SubQuery<String>, String, Character>(lhs,
				LikeOperator.LIKE, rhs, escape);
	}

	public static ConditionalExpression notLike(SubQuery<String> lhs, String rhs) {
		return new BinaryExpression<SubQuery<String>, String>(lhs,
				LikeOperator.NOT_LIKE, rhs);
	}

	public static ConditionalExpression notLike(SubQuery<String> lhs,
			String rhs, Character escape) {
		return new TernaryExpression<SubQuery<String>, String, Character>(lhs,
				LikeOperator.NOT_LIKE, rhs, escape);
	}

	// ------------------- Catch all signatures --------------------------------

	// Catch-all signatures that are necessary because eg. SubQuery<String> has
	// the same erasure as SubQuery<Number>

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression eq(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression neq(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression lt(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression lte(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression gt(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression gte(SubQuery<V> lhs,
			SubQuery<W> rhs) {
		return new BinaryExpression<SubQuery<V>, SubQuery<W>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression eq(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				EqualityOperator.EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression neq(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				EqualityOperator.NOT_EQUALS, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression lt(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				ComparisonOperator.LT, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression lte(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				ComparisonOperator.LTE, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression gt(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				ComparisonOperator.GT, rhs);
	}

	/**
	 * @category type_safety_issue
	 */
	public static <V, W extends V> ConditionalExpression gte(SubQuery<V> lhs,
			AllOrAny<W> rhs) {
		return new BinaryExpression<SubQuery<V>, AllOrAny<W>>(lhs,
				ComparisonOperator.GTE, rhs);
	}

	// -------------------------------------------------------------------------

	// valid IS [NOT] NULL
	public static ConditionalExpression isNull(final Object singleValued) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(singleValued) + " is null";
			}
		};
	}

	public static ConditionalExpression isNotNull(final Object singleValued) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(singleValued)
						+ " is not null";
			}
		};
	}

	// invalid IS [NOT] NULL : return void to attract developers attention
	/**
	 * @category invalid
	 */
	public static void isNull(final Collection<?> multiValued) {
	}

	/**
	 * @category invalid
	 */
	public static void isNotNull(final Collection<?> singleValued) {
	}

	// IS [NOT] EMPTY
	public static ConditionalExpression isEmpty(final Collection<?> multiValued) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(multiValued) + " is empty";
			}
		};
	}

	public static ConditionalExpression isNotEmpty(
			final Collection<?> multiValued) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(multiValued)
						+ " is not empty";
			}
		};
	}

	// state_field_path_expression [NOT] IN (in_item {, in_item}* | subquery)
	// ==> collections are illegal arguments
	// in(Collection) is used for collection_member_declaration stub anyway

	// TODO : see 4.6.8 : should stateField be limited to String, Number & Enum
	public static <W, V extends W> ConditionalExpression in(final V stateField,
			final W inItem, final W... more) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				StringBuilder sb = new StringBuilder(AliasFactory
						.safeStringForm(stateField));
				// TODO : force literals or params
				sb.append(" in (").append(AliasFactory.safeStringForm(inItem));
				if (more != null) {
					for (W addtional : more) {
						sb.append(", ").append(
								AliasFactory.safeStringForm(addtional));
					}
				}
				sb.append(')');
				return sb.toString();
			}
		};
	}

	public static <W, V extends W> ConditionalExpression notIn(
			final V stateField, final W inItem, final W... more) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				StringBuilder sb = new StringBuilder(AliasFactory
						.safeStringForm(stateField));
				// TODO : force literals or params
				sb.append(" not in (").append(
						AliasFactory.safeStringForm(inItem));
				if (more != null) {
					for (W addtional : more) {
						sb.append(", ").append(
								AliasFactory.safeStringForm(addtional));
					}
				}
				sb.append(')');
				return sb.toString();
			}
		};
	}

	// in (subquery)
	public static <W, V extends W> ConditionalExpression in(final V stateField,
			final SubQuery<W> subquery) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				StringBuilder sb = new StringBuilder(AliasFactory
						.safeStringForm(stateField));
				sb.append(" in (").append(subquery.toString()).append(')');
				return sb.toString();
			}
		};
	}

	public static <W, V extends W> ConditionalExpression notIn(
			final V stateField, final SubQuery<W> subquery) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				StringBuilder sb = new StringBuilder(AliasFactory
						.safeStringForm(stateField));
				sb.append(" not in (").append(subquery.toString()).append(')');
				return sb.toString();
			}
		};
	}

	// 4.6.12 [NOT] MEMBER OF
	public static <W> ConditionalExpression memberOf(final W element,
			final Collection<? super W> collection) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(element) + " member of "
						+ AliasFactory.safeStringForm(collection);
			}
		};
	}

	public static <W> ConditionalExpression notMemberOf(final W element,
			final Collection<? super W> collection) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return AliasFactory.safeStringForm(element) + " not member of "
						+ AliasFactory.safeStringForm(collection);
			}
		};
	}

	// [NOT] EXISTS (subquery)
	public static ConditionalExpression exists(final SubQuery<?> subquery) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return "exists (" + subquery.toString() + ")";
			}

		};
	}

	public static ConditionalExpression notExists(final SubQuery<?> subquery) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return "not exists (" + subquery.toString() + ")";
			}

		};
	}

	// ALL, ANY, SOME (subquery)

	public static <W> AllOrAny<W> all(final SubQuery<W> subquery) {
		return new AllOrAny<W>(AllOrAnyOperator.ALL, subquery);
	}

	public static <W> AllOrAny<W> any(final SubQuery<W> subquery) {
		return new AllOrAny<W>(AllOrAnyOperator.ANY, subquery);
	}

	public static <W> AllOrAny<W> some(final SubQuery<W> subquery) {
		return new AllOrAny<W>(AllOrAnyOperator.SOME, subquery);
	}

	// grouping with parentheses
	public static ConditionalExpression paren(final ConditionalExpression what) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return "(" + what.getTargetRepresentation() + ")";
			}
		};
	}

	// negating
	public static ConditionalExpression not(final ConditionalExpression negated) {
		return new ConditionalExpression() {
			@Override
			public String getTargetRepresentation() {
				return "not " + negated.getTargetRepresentation();
			}
		};
	}

}
